from pathlib import Path

import click
import pyperclip

from src import utils
from src.models import Key


@click.group()
def cli():
    utils.get_meta_directory().mkdir(parents=True, exist_ok=True)
    pass


@cli.command()
@click.option(
    '-p', '--master-password',
    prompt=True, hide_input=True, type=str, confirmation_prompt=True)
@click.option(
    '-ic', '--iter-count', prompt=True, type=int, default=1000)
@click.option(
    '-kp', '--keylist-path',
    type=click.Path(), default=utils.get_default_path(),
    prompt=True)
def init(master_password, iter_count, keylist_path):
    if Path(keylist_path).exists():
        click.confirm(
            f'A file already exists at "{keylist_path}". '
            f'Do you want to overwrite it?', abort=True)
    salt = utils.generate_salt()
    hashed = utils.hash_master(master_password, salt, iter_count)
    utils.save_master(keylist_path, hashed, salt, iter_count)
    click.echo(f'Successfully saved new master hash at {keylist_path}!')


@cli.command()
@click.option(
    '-p', '--master-password',
    prompt=True, hide_input=True, type=str)
@click.option(
    '-l', '--label', help='Case-insensitive label for passwords.',
    prompt=True)
@click.option(
    '-kp', '--keylist-path',
    type=click.Path(exists=True),
    default=utils.get_last_used_path(), prompt=True)
def generate(master_password, label, keylist_path):
    keylist = utils.load_keylist(keylist_path)
    abort_if_incorrect_master(master_password, keylist, keylist_path)

    label = label.lower()
    key = keylist.get_key(label)
    if key is None:
        max_length = click.prompt('Max Length', type=int, default=-1)
        max_length = None if max_length <= 0 else max_length

        gen_mode = click.prompt(
            'Gen Mode',
            type=click.Choice(['Base64', 'AlphaNum'], case_sensitive=False),
            default='Base64')
        key = Key(label, gen_mode, max_length)
        keylist.keys.append(key)
    utils.save_keylist(keylist_path, keylist)
    new_password = utils.generate_password(master_password, key)
    pyperclip.copy(new_password)
    click.echo('Password copied to clipboard!')


@cli.command()
@click.option(
    '-p', '--master-password',
    prompt=True, hide_input=True, type=str)
@click.option(
    '-l', '--label', help='Case-insensitive label for passwords.',
    prompt=True)
@click.option(
    '-kp', '--keylist-path',
    type=click.Path(exists=True),
    default=utils.get_last_used_path(), prompt=True)
@click.pass_context
def reset(ctx, master_password, label, keylist_path):
    keylist = utils.load_keylist(keylist_path)
    abort_if_incorrect_master(master_password, keylist, keylist_path)

    label = label.lower()

    keylist.remove_key(label)
    utils.save_keylist(keylist_path, keylist)
    ctx.invoke(generate, master_password, label, keylist_path)


def abort_if_incorrect_master(master_password, keylist, keylist_path):
    if not utils.check_master(master_password, keylist):
        click.echo(
            f'Master password does not match value saved at {keylist_path}.')
        raise click.Abort



import base64
from hashlib import pbkdf2_hmac, sha256
import hmac
import json
from secrets import token_bytes
from pathlib import Path

from src.models import KeyList, Master


# hash and check
def check_master(password, keylist):
    hashed = hash_master(
        password, keylist.master.salt, keylist.master.iter_count)
    return hashed == keylist.master.hashed


def hash_master(password, salt, iter_count):
    # salt is assumed to be base64 encoded
    salt = base64.b64decode(salt)
    hashed = pbkdf2_hmac('sha256', password.encode(), salt, iter_count)
    return base64.b64encode(hashed).decode('utf-8')


def generate_password(master, key):
    hashed = hmac.new(
        master.encode(), msg=key.label.encode(), digestmod=sha256).digest()
    hashed = base64.b64encode(hashed).decode('utf-8')

    if key.gen_mode.lower() == 'alphanum':
        hashed = ''.join([c for c in hashed if c.isalnum()])

    if key.max_length is not None:
        hashed = hashed[:key.max_length]
    return hashed


def generate_salt():
    return base64.b64encode(token_bytes(32)).decode('utf-8')


# I/O
def save_master(path, hashed, salt, iter_count):
    # this implies a clean slate
    master = Master(hashed, salt, iter_count)
    keylist = KeyList(master)
    save_keylist(path, keylist)


def save_keylist(path, keylist):
    with open(path, 'w') as f:
        f.write(str(keylist))
    save_keylist_path(path)


def save_keylist_path(path):
    file = get_meta_directory() / 'keylists.meta.json'
    existing = []

    if file.exists():
        with open(str(file)) as f:
            existing = json.loads(f.read()).get('paths', [])

    with open(str(file), 'w') as f:
        if path in existing:
            existing.remove(path)
        existing.append(path)
        f.write(json.dumps({'paths': existing}))


def get_meta_directory():
    return Path.home() / '.passgen'


def get_keylists_meta_path():
    return get_meta_directory() / 'keylists.meta.json'


def get_default_path():
    return str(get_meta_directory() / 'default.keys.json')


def get_last_used_path():
    file = get_keylists_meta_path()

    if file.exists():
        with open(str(file)) as f:
            existing = json.loads(f.read()).get('paths', [])

        if len(existing) > 0:
            return existing[-1]
    return get_default_path()


def load_keylist(path):
    with open(path) as f:
        json_obj = f.read()
    return KeyList.from_json(json_obj)



import json


class KeyList:
    def __init__(self, master, keys=[], version=1):
        self.master = master
        self.version = version
        self.keys = keys

    @classmethod
    def from_json(cls, json_str):
        obj = json.loads(json_str)
        return cls(
            master=Master.from_dict(obj['Master']),
            keys=[Key.from_dict(k) for k in obj.get('Keys', [])],
            version=obj['Version'])

    def get_key(self, label):
        for key in self.keys:
            if key.label == label:
                return key
        return None

    def remove_key(self, label):
        self.keys = [k for k in self.keys if k.label != label]

    def __str__(self):
        obj = {
            "Master": self.master.to_dict(),
            "Version": self.version,
            "Keys": [k.to_dict() for k in self.keys]
        }
        return json.dumps(obj)


class Key:
    def __init__(self, label, gen_mode='Base64', max_length=None):
        self.label = label
        self.gen_mode = gen_mode
        self.max_length = max_length

    @classmethod
    def from_dict(cls, obj):
        return cls(
            obj['Label'],
            gen_mode=obj['GenMode'],
            max_length=obj['MaxLength'])

    def to_dict(self):
        return {
            "Label": self.label,
            "GenMode": self.gen_mode,
            "MaxLength": self.max_length
        }


class Master:
    def __init__(self, hashed, salt, iter_count):
        self.hashed = hashed
        self.salt = salt
        self.iter_count = iter_count

    @classmethod
    def from_dict(cls, obj):
        return cls(
            obj['Hash'],
            obj['Salt'],
            obj['IterCount'])

    def to_dict(self):
        return {
            "Hash": self.hashed,
            "Salt": self.salt,
            "IterCount": self.iter_count
        }



from setuptools import setup, find_packages

setup(
	name = 'passgen-py',
	packages = find_packages(), 
	version = '0.1.1',
	description = 'Generate Passwords Deterministically based on a Master Password.',
	classifiers = [
		'Development Status :: 3 - Alpha',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3'
	],
	python_requires='>=3.6, <4',
	entry_points={
        'console_scripts': [
            'passgen=src:cli',
        ],
    },
    install_requires=['click', 'pyperclip'],
)
